"""Types relevant to array manipulation and image processing."""

from typing import TypeAlias

import pandas as pd

DF: TypeAlias = pd.DataFrame
DfOrS: TypeAlias = pd.DataFrame | pd.Series  # type: ignore  # pyright 1.1.317



"""Project parameters."""

from pathlib import Path

from boilercore.models import SynchronizedPathsYamlModel
from boilercore.models.fit import Fit
from boilercore.models.geometry import Geometry
from pydantic.v1 import Field

from boilerdaq import get_params_file
from boilerdaq.models.paths import Paths

PARAMS_FILE = get_params_file()


class Params(SynchronizedPathsYamlModel):
    """Project parameters."""

    fit: Fit = Field(default_factory=Fit, description="Parameters for model fit.")
    geometry: Geometry = Field(default_factory=Geometry, description="Geometry.")

    paths: Paths = Field(default_factory=Paths)
    """Project paths."""

    def __init__(self, data_file: Path = PARAMS_FILE, **kwargs):
        """Initialize, propagate paths to the parameters file, and update the schema."""
        super().__init__(data_file, **kwargs)


PARAMS = Params()
"""All project parameters, including paths."""



"""Use heat flux as a control parameter."""

from boilerdaq.daq import (
    Controller,
    FitResult,
    Looper,
    Plotter,
    Result,
    ResultGroup,
    Writer,
    get_result,
)
from boilerdaq.models.params import PARAMS
from boilerdaq.stages import CONTROL_SENSOR_NAME, OUTPUT_LIMITS, RESULTS_PATH
from boilerdaq.stages.controlled import CONTROLLED_RESULTS

SURFACE_TEMP_SETPOINT = 30
PID_GAINS = (12, 0.08, 1)
SURFACE_TEMP = "T_s"
SURFACE_TEMP_UNIT = "C"


def main(surface_temp_setpoint: float = SURFACE_TEMP_SETPOINT):  # noqa: D103
    looper(surface_temp_setpoint).start()


def looper(surface_temp_setpoint: float = SURFACE_TEMP_SETPOINT) -> Looper:  # noqa: D103
    fit_result = FitResult(
        name=SURFACE_TEMP,
        unit=SURFACE_TEMP_UNIT,
        fit=PARAMS.fit,
        models=PARAMS.paths.models,
        results_to_fit=[
            get_result(name=result, results=CONTROLLED_RESULTS)
            for result in "T1cal T2cal T3cal T4cal T5cal".split()
        ],
    )
    results: list[Result] = [*CONTROLLED_RESULTS, fit_result]
    controller = Controller(
        control_result=get_result(name=CONTROL_SENSOR_NAME, results=results),
        feedback_result=fit_result,
        setpoint=surface_temp_setpoint,
        gains=PID_GAINS,
        output_limits=OUTPUT_LIMITS,
    )
    group_dict = dict(
        feedback=SURFACE_TEMP,
        post="T1cal T2cal T3cal T4cal",
        top="T5cal",
        control="V",
        water="Tw1cal Tw2cal Tw3cal",
        pressure="Pcal",
    )
    group = ResultGroup(group_dict, results)
    plotter = Plotter("feedback", group["feedback"], 0, 0)
    plotter.add("post", group["post"], 0, 1)
    plotter.add("top", group["top"], 0, 2)
    plotter.add("control", group["control"], 1, 0)
    plotter.add("water", group["water"], 1, 1)
    plotter.add("pressure", group["pressure"], 1, 2)
    return Looper(Writer(RESULTS_PATH, results), plotter, controller)


if __name__ == "__main__":
    main()



"""Data acquisition for a nucleate pool boiling experimental apparatus."""

from pathlib import Path

PROJECT_PATH = Path()
INSTRUMENT = "USB0::0x0957::0x0807::US25N3188G::0::INSTR"


def get_params_file():  # noqa: D103
    return PROJECT_PATH / "params.yaml"



"""Dummy methods if devices not attached."""


def t_in(_board_num, _channel, _scale, _options=0):
    """Get dummy value for `t_in`."""
    return 1.0


def v_in(_board_num, _channel, _ul_range, _options=0):
    """Get dummy value for `v_in`."""
    return 1.0



"""Paths for this project."""

from pathlib import Path

from boilercore.models import CreatePathsModel
from boilercore.paths import get_package_dir, map_stages
from pydantic.v1 import DirectoryPath, FilePath

import boilerdaq
from boilerdaq import PROJECT_PATH


class Paths(CreatePathsModel):
    """Paths associated with this project."""

    # * Roots
    project: DirectoryPath = PROJECT_PATH
    data: DirectoryPath = project / "data"

    # * Local results
    benchmarks: DirectoryPath = data / "benchmarks"
    results: DirectoryPath = data / "results"
    notes: DirectoryPath = data / "notes"

    # * Git-tracked inputs
    # ! Package
    package: DirectoryPath = get_package_dir(boilerdaq)
    stages: dict[str, FilePath] = map_stages(package / "stages")
    # ! Config
    # Careful, "Config" is a special member of BaseClass
    config: DirectoryPath = data / "config"
    sensors_path = config / "1_sensors.csv"
    scaled_params_path = config / "2_scaled_params.csv"
    flux_params_path = config / "3_flux_params.csv"
    extrap_params_path = config / "4_extrap_params.csv"
    power_supplies_path = config / "0_power_supplies.csv"
    # ! Plotting
    plot_config: DirectoryPath = data / "plotting"
    mpl_base: FilePath = plot_config / "base.mplstyle"
    mpl_hide_title: FilePath = plot_config / "hide_title.mplstyle"

    # * DVC-tracked imports
    models: Path = data / "models"



"""Run the data acquisition loop."""

from boilerdaq.daq import Looper, Writer
from boilerdaq.stages import BASE_RESULTS, PLOTTER, RESULTS_PATH


def main() -> Looper:  # noqa: D103
    return Looper(Writer(RESULTS_PATH, BASE_RESULTS), PLOTTER)


if __name__ == "__main__":
    main().start()



"""Prepare the feedback-controlled data acquisition loop."""

from boilerdaq import INSTRUMENT
from boilerdaq.daq import Plotter, PowerParam, PowerResult, ResultGroup, Writer
from boilerdaq.models.params import PARAMS
from boilerdaq.stages import BASE_RESULTS, CURRENT_LIMIT, GROUP_DICT, RESULTS_PATH

# Get power supply values
all_power_supplies = PowerParam.get(PARAMS.paths.power_supplies_path)
power_results = []
for power_supply in all_power_supplies:
    result = PowerResult(power_supply, INSTRUMENT, CURRENT_LIMIT)
    power_results.append(result)
CONTROLLED_RESULTS = power_results + BASE_RESULTS

# Create the writer
writer = Writer(RESULTS_PATH, CONTROLLED_RESULTS)
group = ResultGroup(GROUP_DICT, CONTROLLED_RESULTS)

# Create the plotter and add groups of curves to different plot regions
PLOTTER = Plotter("base", group["base"], 0, 0)
PLOTTER.add("post", group["post"], 0, 1)
PLOTTER.add("top", group["top"], 0, 2)
PLOTTER.add("water", group["water"], 1, 0)
PLOTTER.add("pressure", group["pressure"], 1, 1)
PLOTTER.add("flux", group["flux"], 1, 2)



"""Calibrate thermocouples."""

from boilerdaq.daq import Looper, Plotter, ResultGroup, Writer
from boilerdaq.stages import READINGS, RESULTS_PATH, SCALED_RESULTS

# Build list of sensor groups, grouped by name
GROUP_DICT = dict(
    uncalibrated="T0 T1 T2 T3 T4 T5 Tw1 Tw2 Tw3 Tw4",
    calibrated="T0cal T1cal T2cal T3cal T4cal T5cal Tw1cal Tw2cal Tw3cal Tw4cal",
    pressure_volts="P",
    pressure="Pcal",
)


def main() -> Looper:  # noqa: D103
    # Combine calculated results into one list
    results = READINGS + SCALED_RESULTS
    # Start writing
    writer = Writer(RESULTS_PATH, results)
    group = ResultGroup(GROUP_DICT, results)
    # Add groups of curves to different plot regions
    plotter = Plotter("uncalibrated", group["uncalibrated"], 0, 0)
    plotter.add("calibrated", group["calibrated"], 0, 1)
    plotter.add("pressure_volts", group["pressure_volts"], 1, 0)
    plotter.add("pressure", group["pressure"], 1, 1)
    # Start the write and plot loops
    return Looper(writer, plotter)


if __name__ == "__main__":
    main().start()



"""Run the data acquisition and control loop."""

from boilerdaq.daq import Controller, Looper, Writer, get_result
from boilerdaq.stages import CONTROL_SENSOR_NAME, OUTPUT_LIMITS, RESULTS_PATH
from boilerdaq.stages.controlled import CONTROLLED_RESULTS, PLOTTER

TEMP_SETPOINT = 30
TEMP_FEEDBACK_GAINS = (12, 0.08, 1)
TEMP_FEEDBACK_SENSOR_NAME = "T0cal"


def main() -> Looper:  # noqa: D103
    controller = Controller(
        get_result(CONTROL_SENSOR_NAME, CONTROLLED_RESULTS),  # type: ignore
        get_result(TEMP_FEEDBACK_SENSOR_NAME, CONTROLLED_RESULTS),
        TEMP_SETPOINT,
        TEMP_FEEDBACK_GAINS,
        OUTPUT_LIMITS,
    )
    return Looper(Writer(RESULTS_PATH, CONTROLLED_RESULTS), PLOTTER, controller)


if __name__ == "__main__":
    main().start()



"""Prepare the data acquisition loop."""

from boilerdaq.daq import (
    ExtrapParam,
    ExtrapResult,
    Flux,
    FluxParam,
    Plotter,
    Reading,
    ResultGroup,
    ScaledParam,
    ScaledResult,
    Sensor,
)
from boilerdaq.models.params import PARAMS

RESULTS_PATH = PARAMS.paths.results / "results.csv"
CURRENT_LIMIT = 4
CONTROL_SENSOR_NAME = "V"
OUTPUT_LIMITS = (0, 300)

# Build list of sensor groups, grouped by name
GROUP_DICT = dict(
    base="T0cal",
    post="T1cal T2cal T3cal T4cal",
    top="T5cal T6ext",
    water="Tw1cal Tw2cal Tw3cal",
    pressure="Pcal",
    flux="Q12 Q23 Q34 Q45",
)

# Get all readings
all_sensors = Sensor.get(PARAMS.paths.sensors_path)
READINGS = []
for sensor in all_sensors:
    reading = Reading(sensor)
    READINGS.append(reading)

# Get scaled parameters
scaled_params = ScaledParam.get(PARAMS.paths.scaled_params_path)
# Get scaled results
SCALED_RESULTS = []
for param in scaled_params:
    result = ScaledResult(param, READINGS)
    SCALED_RESULTS.append(result)

# Get flux parameters
flux_params = FluxParam.get(PARAMS.paths.flux_params_path)
# Get fluxes
fluxes = []
for param in flux_params:
    result = Flux(param, SCALED_RESULTS)
    fluxes.append(result)

# Get extrapolation parameters
extrap_params = ExtrapParam.get(PARAMS.paths.extrap_params_path)
# Get extrapolated results
extrap_results = []
for param in extrap_params:
    result = ExtrapResult(param, SCALED_RESULTS + fluxes)
    extrap_results.append(result)

BASE_RESULTS = READINGS + SCALED_RESULTS + fluxes + extrap_results

# Build list of sensor groups, grouped by name
group_dict = dict(
    base="T0cal",
    post="T1cal T2cal T3cal T4cal",
    top="T5cal T6ext",
    water="Tw1cal Tw2cal Tw3cal",
    pressure="Pcal",
    flux="Q12 Q23 Q34 Q45",
)
group = ResultGroup(group_dict, BASE_RESULTS)

# Create the plotter and add groups of curves to different plot regions
PLOTTER = Plotter("base", group["base"], 0, 0)
PLOTTER.add("post", group["post"], 0, 1)
PLOTTER.add("top", group["top"], 0, 2)
PLOTTER.add("water", group["water"], 1, 0)
PLOTTER.add("pressure", group["pressure"], 1, 1)
PLOTTER.add("flux", group["flux"], 1, 2)



"""Shim that wraps some methods from UL for use in Python in Linux.

MIT License

Copyright (c) 2018 Measurement Computing

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# pyright: basic

from ctypes import CDLL, POINTER, byref, c_float, c_int, create_string_buffer
from ctypes.util import find_library
from enum import IntFlag
from sys import platform

if platform.startswith("darwin"):
    lib_file_name = "libuldaq.dylib"
else:
    lib_file_name = "libuldaq.so"

lib_file_path = find_library("uldaq")
if lib_file_path is None:
    lib_file_path = lib_file_name
lib = CDLL(lib_file_path)
lib.ulTIn.argtypes = [c_int, c_int, c_int, POINTER(c_float), c_int]


def _check_err(errcode):
    if errcode:
        raise ULError(errcode)


class ULError(Exception):  # noqa: D101
    def __init__(self, errorcode):
        super().__init__()
        self.errorcode = errorcode
        self.message = get_err_msg(errorcode)

    def __str__(self):
        return "Error " + str(self.errorcode) + ": " + self.message


_ERRSTRLEN = 256


def get_err_msg(error_code):
    """Return the error message associated with an error code.

    It is usually unnecessary to use
    this method externally, since ULError errors raised by this module assign the value to the
    message property. Some functions do return an error code in cases where it is used as a
    warning. Call this function to convert the returned error code to a descriptive error message.

    Parameters
    ----------
    error_code
        The error code that is returned by some functions in this library.

    Returns
    -------
    string
        The error message associated with the given error_code
    """
    msg = create_string_buffer(_ERRSTRLEN)
    _check_err(lib.ulGetErrMsg(error_code, msg))
    return msg.value.decode("utf-8")


class TInOptions(IntFlag):  # noqa: D101
    FILTER = 0x0000  # Filter thermocouple inputs
    NOFILTER = 0x0400  # Disable filtering for thermocouple

    WAITFORNEWDATA = 0x2000  # Wait for new data to become available


def t_in(board_num, channel, scale, options=TInOptions.FILTER):
    """Read an analog input channel.

    Linearize it according to the selected temperature sensor type, if required, and
    return the temperature in units determined by the scale parameter.

    Parameters
    ----------
    board_num : int
        The number associated with the board when it was installed with InstaCal or created
        with :func:`.create_daq_device`.
    channel : int
        Input channel to read.
    scale : TempScale
        Specifies the temperature scale that the input will be converted to
    options : TInOptions, optional
        Flags that control various options. Refer to the constants in the "options parameter values"
        section below.

    Returns
    -------
    float
        The temperature value


    .. table:: **options parameter values**

        ==============  ===========================================================================
        FILTER          When selected, a smoothing function is applied to temperature readings,
                        very much like the electrical smoothing inherent in all hand held
                        temperature sensor instruments. This is the default. When selected, 10
                        samples are read from the specified channel and averaged. The average is
                        the reading returned. Averaging removes normally distributed signal line
                        noise.
        --------------  ---------------------------------------------------------------------------
        NOFILTER        If you use the NOFILTER option then the readings will not be smoothed and
                        you will see a scattering of readings around a mean.
        --------------  ---------------------------------------------------------------------------
        WAITFORNEWDATA  Waits for new data to become available.
        ==============  ===========================================================================



    Notes
    -----
    **Scale options**

    - Specify the TempScale.NOSCALE to the scale parameter to retrieve raw data from the device.
      When TempScale.NOSCALE is specified, calibrated data is returned, although a cold junction
      compensation (CJC) correction factor is not applied to the returned values.

    - Specify the :const:`~pyulmcculwms.TempScale.VOLTS` option to read the voltage input of a
      thermocouple.

    Refer to board-specific information in the Universal Library User's Guide to determine if your
    hardware supports these options.
    """
    temp_value = c_float()
    _check_err(lib.ulTIn(board_num, channel, scale, byref(temp_value), options))
    return temp_value.value


def v_in(board_num, channel, ul_range, options=0):
    """Read an A/D input channel, return a voltage value.

    If the specified A/D board has programmable gain, then this function sets the gain
    to the specified range.

    Parameters
    ----------
    board_num : int
        The number associated with the board when it was installed with InstaCal or
        created with :func:`.create_daq_device`.
    channel : int
        A/D channel number. The maximum allowable channel depends on which type of A/D
        board is being used. For boards with both single-ended and differential inputs,
        the maximum allowable channel number also depends on how the board is
        configured.
    ul_range : ULRange
        A/D range code. If the board has a programmable gain, it will be set according
        to this parameter value. Keep in mind that some A/D boards have a programmable
        gain feature, and others set the gain via switches on the board. In either case,
        the range that the board is configured for must be passed to this function.
        Refer to board specific information for a list of the supported A/D ranges of
        each board.
    options : int, optional
        Reserved for future use

    Returns
    -------
    float
        The value in volts of the A/D sample
    """
    data_value = c_float()
    _check_err(lib.ulVIn(board_num, channel, ul_range, byref(data_value), options))
    return data_value.value



"""Simple test with one sensor."""

from boilerdaq.daq import Looper, Plotter, ResultGroup, Writer
from boilerdaq.stages import READINGS, RESULTS_PATH

GROUP_DICT = dict(base="T1")


def main() -> Looper:  # noqa: D103
    writer = Writer(RESULTS_PATH, READINGS)
    group = ResultGroup(GROUP_DICT, READINGS)
    plotter = Plotter("base", group["base"], 0, 0)
    return Looper(writer, plotter)


if __name__ == "__main__":
    main().start()



"""Set the power supply go a given voltage."""

from pyvisa.resources import MessageBasedResource

from boilerdaq.daq import get_result
from boilerdaq.stages.controlled import CONTROLLED_RESULTS

CURRENT_LIMIT = 4
VOLTAGE = 30


def main():  # noqa: D103
    result = get_result("V", CONTROLLED_RESULTS)
    result.open()  # type: ignore
    instrument: MessageBasedResource = result.instrument  # type: ignore
    instrument.write(f"source:current {CURRENT_LIMIT}")
    instrument.write(f"source:voltage {VOLTAGE}")
    instrument.write("output:state on")
    input("Press ENTER to turn off the power supply... ")
    instrument.write("output:state off")
    instrument.write("source:current 0")
    instrument.close()


if __name__ == "__main__":
    main()



"""Data acquisition functions."""

from collections import UserDict, deque
from contextlib import suppress
from csv import DictReader, DictWriter
from datetime import datetime, timedelta
from pathlib import Path
from tempfile import NamedTemporaryFile
from textwrap import dedent
from typing import Any, NamedTuple, Self
from warnings import warn

from boilercore.fits import fit_from_params
from boilercore.modelfun import get_model
from boilercore.models.fit import Fit
from boilercore.models.geometry import GEOMETRY
from boilercore.types import Rod
from pyqtgraph import (
    DateAxisItem,
    GraphicsLayoutWidget,
    PlotCurveItem,
    intColor,
    mkQApp,
    setConfigOptions,
)
from PySide6.QtCore import Qt, QTimer, Signal
from PySide6.QtGui import QKeyEvent
from pyvisa import ResourceManager, VisaIOError
from pyvisa.resources import MessageBasedResource
from simple_pid import PID

try:
    try:
        from mcculw.enums import InterfaceType
        from mcculw.ul import ULError, get_daq_device_inventory, t_in, v_in
    except NameError:
        from uldaq import InterfaceType, get_daq_device_inventory

        from boilerdaq.shim import t_in, v_in

    if not get_daq_device_inventory(InterfaceType.USB):  # pyright: ignore[reportArgumentType]
        from boilerdaq.dummy import t_in, v_in
except FileNotFoundError:
    from boilerdaq.dummy import t_in, v_in


setConfigOptions(antialias=True)

PLOT_HISTORY_DURATION = 5  # (min)
"""Duration of plot history."""
POLLING_INTERVAL = 2000
"""Minimum time between sampling cycles in milliseconds."""
PLOT_HISTORY_LENGTH = PLOT_HISTORY_DURATION * 60 * 1000 // POLLING_INTERVAL
"""Length of plot history in number of samples."""


class Sensor(NamedTuple):
    """Sensor parameters.

    Parameters
    ----------
    name: str
        Name of the sensor.
    board: int
        Which board the sensor belongs to.
    channel: int
        The channel pointing to this sensor on the board.
    reading: int
        The sensor type, either "Temperature" or "Voltage".
    unit: str
        The unit type for values reported by the board.
    """

    name: str
    board: int
    channel: int
    reading: str
    unit: str

    @classmethod
    def get(cls, path: Path) -> list[Self]:
        """Process a CSV file at ``path``, returning a ``List`` of ``Sensor``."""
        sensors = []
        with path.open(encoding="utf-8") as csv_file:
            reader = DictReader(csv_file)
            sensors.extend(
                cls(
                    row["name"],
                    int(row["board"]),
                    int(row["channel"]),
                    row["reading"],
                    row["unit"],
                )
                for row in reader
            )
        return sensors


class Param(NamedTuple):
    """A generic parameter.

    Parameters
    ----------
    name: str
        The name of the parameter.
    unit:
        The unit of the parameter.
    """

    name: str
    unit: str


class ScaledParam(NamedTuple):
    """Parameters for scalar modification of a sensor.

    Parameters
    ----------
    name: str
        Name of the scaled value.
    unscaled_sensor: str
        Name of the sensor to be scaled.
    scale: float
        The scale to apply.
    offset: float
        The offset to apply.
    unit: str
        The unit type after scaling.
    """

    name: str
    unscaled_sensor: str
    scale: float
    offset: float
    unit: str

    @classmethod
    def get(cls, path: Path) -> list[Self]:
        """Process a CSV file at ``path``, returning a ``List`` of ``ScaledParam``."""
        params = []
        with path.open(encoding="utf-8") as csv_file:
            reader = DictReader(csv_file)
            params.extend(
                cls(
                    row["name"],
                    row["unscaled_sensor"],
                    float(row["scale"]),
                    float(row["offset"]),
                    str(row["unit"]),
                )
                for row in reader
            )
        return params


class FluxParam(NamedTuple):
    """Parameters for the flux between two sensors.

    Parameters
    ----------
    name: str
        The name of the flux.
    origin_sensor: str
        The name of the sensor at the origin.
    distant_sensor: str
        The name of the sensor not at the origin.
    conductivity: float
        The conductivity of the path between the sensors.
    length: float
        The length of the path between the sensors.
    unit: str
        The unit type of the flux.
    """

    name: str
    origin_sensor: str
    distant_sensor: str
    conductivity: float
    length: float
    unit: str

    @classmethod
    def get(cls, path: Path) -> list[Self]:
        """Process a CSV file at ``path``, returning a ``List`` of ``FluxParam``."""
        params = []
        with path.open(encoding="utf-8") as csv_file:
            reader = DictReader(csv_file)
            params.extend(
                cls(
                    row["name"],
                    row["origin_sensor"],
                    row["distant_sensor"],
                    float(row["conductivity"]),
                    float(row["length"]),
                    row["unit"],
                )
                for row in reader
            )
        return params


class ExtrapParam(NamedTuple):
    """Parameters for extrapolation from two sensors and a flux to a point of interest.

    Parameters
    ----------
    name: str
        The name of the extrapolation.
    origin_sensor: str
        The name of the sensor at the origin.
    distant_sensor: str
        The name of the sensor not at the origin.
    conductivity: float
        The conductivity of the path from ``distant_sensor`` to the point of interest.
    length: float
        The length of the path from ``distant_sensor`` to the point of interest.
    unit: str
        The unit type of the extrapolation.
    """

    name: str
    origin_sensor: str
    flux: str
    conductivity: float
    length: float
    unit: str

    @classmethod
    def get(cls, path: Path) -> list[Self]:
        """Process a CSV file at ``path``, returning a ``List`` of ``ExtrapParam``."""
        params = []
        with path.open(encoding="utf-8") as csv_file:
            reader = DictReader(csv_file)
            params.extend(
                cls(
                    row["name"],
                    row["origin_sensor"],
                    row["flux"],
                    float(row["conductivity"]),
                    float(row["length"]),
                    row["unit"],
                )
                for row in reader
            )
        return params


class PowerParam(NamedTuple):
    """Parameters for power supplies.

    Parameters
    ----------
    name: str
        The name of the power supply parameter.
    unit:
        The unit of the power supply parameter.
    """

    name: str
    unit: str

    @classmethod
    def get(cls, path: Path) -> list[Self]:
        """Process a CSV file at ``path``, returning a ``List`` of ``PowerParam``."""
        power_supplies = []
        with path.open(encoding="utf-8") as csv_file:
            reader = DictReader(csv_file)
            power_supplies.extend(cls(row["name"], row["unit"]) for row in reader)
        return power_supplies


class Result:
    """A result.

    Attributes
    ----------
    source: str
        The source of the result.
    value: float
        The value of the result.
    time: float
        The time that the result was taken, with the oldest result at zero.
    history: Deque[float]
        Previous values resulting from the source.
    """

    def __init__(self):
        self.source: Sensor = None  # type: ignore
        self.value: float = None  # type: ignore
        self.history: deque[float] = deque(maxlen=PLOT_HISTORY_LENGTH)

    def update(self):
        """Update the result."""
        self.history.append(self.value)


def get_result(name: str, results: list[Result]) -> Result:
    """Get a result or results by the source name."""
    result_names = [result.source.name for result in results]
    i = result_names.index(name)
    return results[i]


UNIT_TYPES = {"C": 0, "F": 1, "K": 2, "V": 5}


class Reading(Result):
    """A reading directly from a sensor.

    Parameters
    ----------
    sensor: Sensor
        The sensor parameters used to get a result.

    Attributes
    ----------
    unit_types: Dict[str: int]
        Enumeration of unit types supported by the board on which the sensor resides.
    """

    def __init__(self, sensor: Sensor):
        super().__init__()
        self.source = sensor

    def update(self):
        """Update the result."""
        if self.source.reading == "temperature":
            try:
                unit_int = UNIT_TYPES[self.source.unit]
                self.value = t_in(self.source.board, self.source.channel, unit_int)
            except ULError:  # type: ignore
                self.value = 0
        elif self.source.reading == "voltage":
            self.value = v_in(self.source.board, self.source.channel, 0)
        super().update()


class ScaledResult(Result):
    """A scaled result.

    Parameters
    ----------
    scaled_param: ScaledParam
        The parameters for obtaining a scaled result.
    results: List[Result]
        A list of results containing the source to be scaled.

    Attributes
    ----------
    unscaled_result: Result
        The unscaled result.
    """

    def __init__(self, scaled_param: ScaledParam, results: list[Result]):
        super().__init__()
        self.source: ScaledParam = scaled_param  # type: ignore
        self.unscaled_result = get_result(scaled_param.unscaled_sensor, results)

    def update(self):
        """Update the result."""
        self.value = self.unscaled_result.value * self.source.scale + self.source.offset  # type: ignore
        super().update()


class Flux(Result):
    """A flux result.

    Parameters
    ----------
    flux_param: FluxParam
        The parameters for obtaining a flux result.
    results: List[Result]
        A list of results containing the source to be scaled.

    Attributes
    ----------
    origin_result: Result
        The result of the source at the origin.
    distant_result: Result
        The result of the source not at the origin.
    """

    def __init__(self, flux_param: FluxParam, results: list[Result]):
        super().__init__()
        self.source: FluxParam = flux_param  # type: ignore
        self.origin_result = get_result(flux_param.origin_sensor, results)
        self.distant_result = get_result(flux_param.distant_sensor, results)

    def update(self):
        """Update the result."""
        self.value = (
            self.source.conductivity
            / self.source.length
            * (self.origin_result.value - self.distant_result.value)  # type: ignore
        )
        super().update()


class ExtrapResult(Result):
    """An extrapolated result.

    Parameters
    ----------
    extrap_param: ExtrapParam
        The parameters for obtaining an extrapolated result.
    results: List[Result]
        A list of results containing the source to be scaled.

    Attributes
    ----------
    origin_result: Result
        The result of the source at the origin.
    distant_result: Result
        The result of the source not at the origin.
    """

    def __init__(self, extrap_param: ExtrapParam, results: list[Result]):
        super().__init__()
        self.source: ExtrapParam = extrap_param  # type: ignore
        self.origin_result = get_result(extrap_param.origin_sensor, results)
        self.flux_result = get_result(extrap_param.flux, results)

    def update(self):
        """Update the result."""
        self.value = self.origin_result.value - (
            self.flux_result.value * self.source.length / self.source.conductivity
        )
        super().update()


class FitResult(Result):
    """A result from a model fit."""

    def __init__(
        self,
        name: str,
        unit: str,
        fit: Fit,
        models: Path,
        results_to_fit: list[Result],
        rod: Rod = "R",
    ):
        super().__init__()
        self.name = name
        self.unit = unit
        self.fit = fit
        self.source = Param(name, unit)  # type: ignore
        self.results_to_fit = results_to_fit
        self.model, _ = get_model(models)
        self.x = GEOMETRY.rods[rod]

    def update(self):
        """Update the result."""
        fits, _errors = fit_from_params(
            model=self.model,
            params=self.fit,
            x=self.x,
            y=[result.value for result in self.results_to_fit],
        )
        self.value = fits[self.name]
        super().update()


def open_instrument(name: str) -> MessageBasedResource:
    """Open an instrument by name, falling back to a simulated instrument."""
    rm = ResourceManager()
    if rm.list_resources(name):
        return rm.open_resource(name, read_termination="\n", write_termination="\n")  # pyright: ignore[reportReturnType]
    with NamedTemporaryFile(encoding="utf-8", mode="w", delete=False) as f:
        # sourcery skip: extract-method
        f.write(
            dedent(rf"""
                spec: "1.0"
                devices:
                    device:
                        eom:
                            USB INSTR:
                                q: "\n"
                                r: "\n"
                        dialogues:
                            - q: "measure:voltage?"
                              r: "1"
                            - q: "measure:current?"
                              r: "1"
                resources:
                    {name}:
                        device: "device"
                """).strip()
            + "\n"
        )
        f.close()
        path = Path(f.name)
        inst = ResourceManager(f"{path.as_posix()}@sim").open_resource(
            name, read_termination="\n", write_termination="\n"
        )
        path.unlink()
        return inst  # pyright: ignore[reportReturnType]


class PowerResult(Result):
    """A result from a power supply.

    Parameters
    ----------
    power_param: PowerParam
        The parameters for obtaining a result from the power supply.
    instrument
        The VISA instrument from which to obtain the result.
    current_limit: float
        The current limit to be set.
    """

    def __init__(self, power_param: PowerParam, instrument: str, current_limit: float):
        super().__init__()
        self.source: PowerParam = power_param  # type: ignore
        self.instrument_name: str = instrument
        self.instrument: MessageBasedResource | None = None  # type: ignore
        self.current_limit = current_limit

    def open(self):
        """Open the instrument."""
        self.instrument = open_instrument(self.instrument_name)
        self.instrument.write("output:state on")  # type: ignore
        self.instrument.write(f"source:current {self.current_limit}")

    def close(self):
        """Close the instrument."""
        if not self.instrument:
            return
        # We don't actually write zero here because the PID will get `starting_output`
        # by querying voltage next time the supply is turned on.
        self.instrument.write("output:state off")  # type: ignore
        self.instrument.close()  # type: ignore
        self.instrument = None

    def one_shot(self):
        """Take a single measurement."""
        self.open()
        self.update()
        self.close()

    def update(self):
        """Update the result."""
        try:
            if self.source.name == "V":
                self.value = float(self.instrument.query("measure:voltage?"))  # type: ignore
            elif self.source.name == "I":
                self.value = float(self.instrument.query("measure:current?"))  # type: ignore
        except VisaIOError as exc:
            warn(str(exc), stacklevel=2)
        super().update()

    def write(self, value):
        """Write a value back to the instrument."""
        try:
            if self.source.name == "V":
                self.instrument.write(f"source:voltage {value!s}")  # type: ignore
            elif self.source.name == "I":
                self.instrument.write(f"source:current {value!s}")  # type: ignore
        except VisaIOError as exc:
            warn(str(exc), stacklevel=2)


class ResultGroup(UserDict[str, list[Result]]):
    """A group of results.

    Parameters
    ----------
    group_dict
        Dictionary of result groupings.
    results: List[Result]
        List of results containing the results to be grouped.
    """

    def __init__(self, group_dict: dict[str, str], results: list[Result]):
        super().__init__()
        for key, val in group_dict.items():
            result_names = val.split()
            filtered_results = []
            for name in result_names:
                result = get_result(name, results)
                filtered_results.append(result)
            self[key] = filtered_results  # type: ignore


class Controller:
    """A PID controller.

    Parameters
    ----------
    control_result: Result
        The result to control based on feedback.
    feedback_result: Result
        The result to get feedback from.
    setpoint: float
        The value that the feedback should be coerced to through PID control.
    gains: List[float]
        List of the proportional, integral, and derivative gains of the PID controller.
    output_limits: Tuple[float, float]
        Limits of the PID controller.
    start_delay: float
        Time to wait before activating PID.

    Attributes
    ----------
    pid: PID
        The PID controller.
    """

    def __init__(
        self,
        control_result: Result,
        feedback_result: Result,
        setpoint: float,
        gains: tuple[float, float, float],
        output_limits: tuple[float, float],
    ):
        self.control_result: PowerResult = control_result  # type: ignore
        self.feedback_result: Result = feedback_result
        self.pid = PID(
            *gains,
            setpoint=setpoint,
            sample_time=None,  # Controlled by POLLING_INTERVAL
            output_limits=output_limits,
            starting_output=self.control_result.value,
        )
        self.feedback_value = self.feedback_result.value

    def start(self):
        """Start the controller."""
        self.control_result.open()

    def close(self):
        """Close the instrument to be controlled."""
        self.control_result.close()

    def update(self):
        """Update the PID controller."""
        self.feedback_value = self.feedback_result.value
        control_value = self.pid(self.feedback_value)
        self.control_result.write(control_value)


class Writer:
    """A CSV file writer.

    Parameters
    ----------
    path: Path
        Base name of the first results CSV to be written (e.g. `results.csv`). The ISO
        time of creation of the file will be appended to the provided path (e.g.
        `results_yyyy_mm_ddThh-mm-ss.csv`).
    results: List[Result]
        The first list of results to be written to a file.

    Attributes
    ----------
    paths: List[str]
        Base names of multiple results CSVs to be written to.
    result_groups: List[List[Result]]
        Groups of results to be written to each of the names in `paths`.
    fieldname_groups: List[str]
        Groups of fieldnames to be written to each of the names in `paths`.
    time: datetime
        The time that the last value was taken.
    """

    def __init__(self, path: Path, results: list[Result]):
        self.paths: list[Path] = []
        self.results: list[Any] = []
        self.result_groups: list[list[Result]] = []
        self.fieldname_groups: list[list[str]] = []
        self.time: datetime = datetime.now()  # type: ignore
        self.add(path, results)

    def add(self, path: Path, results: list[Result]):
        """Add a CSV file to be written to and a set of results to write to it.

        Parameters
        ----------
        path: Path
            Base name of additional results CSVs to be written (e.g. `results.csv`). The
            ISO time of creation of the file will be appended to the provided path (e.g.
            `results_yyyy_mm_ddThh-mm-ss.csv`).
        results: List[Result]
            Additonal list of results to be written to a file.
        """
        # The ":" in ISO time strings is not supported by filenames
        file_time = self.time.isoformat(timespec="seconds").replace(":", "-")  # type: ignore
        path = path.with_stem(f"{path.name}_{file_time}")
        # Compose the fieldnames and first row of values
        sources = [f"{result.source.name} ({result.source.unit})" for result in results]
        fieldnames = ["time", *sources]
        for result in results:
            if isinstance(result, PowerResult):
                result.one_shot()
            else:
                result.update()
            with suppress(AttributeError):
                result.history.extend([result.value] * (PLOT_HISTORY_LENGTH - 1))
        values = [self.time.isoformat()] + [result.value for result in results]  # type: ignore
        to_write = dict(zip(fieldnames, values, strict=True))

        # Create the CSV, writing the header and the first row of values
        with path.open("w", newline="", encoding="utf-8") as csv_file:
            csv_writer = DictWriter(csv_file, fieldnames=fieldnames)
            csv_writer.writeheader()
            csv_writer.writerow(to_write)

        # Record the file and results for writing additional rows later.
        self.paths.append(path)
        self.results.extend(results)
        self.result_groups.append(results)
        self.fieldname_groups.append(fieldnames)

    def start(self):
        """Start the writer."""
        for result in self.results:
            if isinstance(result, PowerResult):
                result.open()

    def update(self):
        """Update results and write the new data to CSV."""
        self.time: str = datetime.now().isoformat()
        for results in self.result_groups:
            for result in results:
                result.update()
        self.write()

    def write(self):
        """Write data to CSV."""
        for path, results, fieldnames in zip(
            self.paths, self.result_groups, self.fieldname_groups, strict=True
        ):
            values = [self.time] + [result.value for result in results]
            to_write = dict(zip(fieldnames, values, strict=True))

            with path.open("a", newline="", encoding="utf-8") as csv_file:
                csv_writer = DictWriter(csv_file, fieldnames=fieldnames)
                csv_writer.writerow(to_write)


class GraphicsLayoutWidgetWithKeySignal(GraphicsLayoutWidget):
    """Emit key signals on `key_signal`."""

    key_signal = Signal(QKeyEvent)

    def keyPressEvent(self, ev: QKeyEvent):  # noqa: N802
        """Handle keypresses."""
        super().keyPressEvent(ev)
        self.key_signal.emit(ev)


class Plotter:
    """A plotter for data.

    Parameters
    ----------
    title: str
        The title of the first plot.
    results: List[Result]
        The results to plot.
    row: int = 0
        The window row to place the first plot.
    col: int = 0
        The window column to place the first plot.

    Attributes
    ----------
    all_results: List[Result]
    all_curves: List[PlotCurveItem]
    all_histories: List[Deque]
    time: List[int]
    """

    def __init__(self, title: str, results: list[Result], row: int = 0, col: int = 0):
        self.window = GraphicsLayoutWidgetWithKeySignal()
        self.app = mkQApp()
        self.window.key_signal.connect(self.keyPressEvent)
        self.all_results: list[Result] = []
        self.all_curves: list[PlotCurveItem] = []
        self.all_histories: list[deque[float]] = []
        self.time: deque[float] = deque(maxlen=PLOT_HISTORY_LENGTH)
        current_time = datetime.now()
        self.time.extendleft(
            (current_time - i * timedelta(milliseconds=POLLING_INTERVAL)).timestamp()
            for i in range(PLOT_HISTORY_LENGTH)
        )
        self.add(title, results, row, col)

    def keyPressEvent(self, ev: QKeyEvent):  # noqa: N802
        """Handle quit events and propagate keypresses to image views."""
        if ev.key() in (Qt.Key.Key_Escape, Qt.Key.Key_Q, Qt.Key.Key_Enter):
            self.app.closeAllWindows()
            ev.accept()

    def add(self, title: str, results: list[Result], row: int, col: int):
        """Plot results to a new pane in the plot window.

        Parameters
        ----------
        title: str
            The title of an additional plot.
        results: List[Result]
            The results to plot.
        row: int = 0
            The window row to place an additional plot.
        col: int = 0
            The window column to place an additional plot.
        """
        plot = self.window.addPlot(row, col)
        plot.addLegend()
        plot.setAxisItems({"bottom": DateAxisItem()})
        plot.setLabel("left", units=results[0].source.unit)
        plot.setTitle(title)
        self.all_results.extend(results)
        histories = [result.history for result in results]
        self.all_histories.extend(histories)
        names = [result.source.name for result in results]
        for i, (history, name) in enumerate(zip(histories, names, strict=True)):
            curve = plot.plot(self.time, history, pen=intColor(i), name=name)
            self.all_curves.append(curve)

    def update(self):
        """Update plots."""
        self.time.append(datetime.now().timestamp())
        for curve, history in zip(self.all_curves, self.all_histories, strict=True):
            curve.setData(self.time, history)


class Looper:
    """Handles threads for plotting, writing, and control.

    Parameters
    ----------
    writer: Writer
        The writer.
    plotter: Plotter
        The plotter.
    controller: Optional[Controller]
        The controller.

    Attributes
    ----------
    plot_window_open: bool
        Whether the plot window is currently open.
    """

    def __init__(
        self, writer: Writer, plotter: Plotter, controller: Controller | None = None
    ):
        self.writer = writer
        self.plotter = plotter
        self.controller: Controller = controller or None  # type: ignore

    def start(self):
        """Start the write/control thread and plot on the main thread."""
        self.writer.start()
        if self.controller:
            self.controller.start()
        timer = QTimer()
        timer.timeout.connect(self.plot_control if self.controller else self.plot)
        timer.start(POLLING_INTERVAL)
        self.plotter.window.show()
        self.plotter.app.exec()
        self.plotter.app.quit()
        if self.controller:
            self.controller.close()

    def plot(self):
        """Plot function."""
        self.plotter.update()
        self.writer.update()

    def plot_control(self):
        """Plot and control."""
        self.plotter.update()
        self.writer.update()
        self.controller.update()
